using System;
using System.IO;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Tests {
  public class ImapSimulatedServer
  {
    public IPEndPoint ServerEndPoint {
      get { return server.LocalEndpoint as IPEndPoint; } 
    }

    public ImapSimulatedServer()
    {
      server = new TcpListener(new IPEndPoint(IPAddress.Loopback, 0));
      server.Start();
    }

    public void Start()
    {
      thread = new Thread(ServerThread);
      thread.Start();
    }

    public void Stop()
    {
      thread.Abort();
    }

    public void EnqueueResponse(string response)
    {
      responseQueue.Enqueue(response);
    }

    public string DequeueRequest()
    {
      for (var count = 0; count < 10; count++) {
        if (requestQueue.Count != 0)
          break;
        else
          Thread.Sleep(50);
      }

      var req = string.Empty;

      // XXX
      lock ((requestQueue as System.Collections.ICollection).SyncRoot) {
        while (0 < requestQueue.Count) {
          req += requestQueue.Dequeue();
          if (req.EndsWith("\r\n"))
            return req;
        }
      }

      return req;
    }

    private void ServerThread()
    {
      var client = server.AcceptTcpClient();

      Console.WriteLine("accepted");

      try {
        var stream = client.GetStream();

        for (;;) {
          if (responseQueue.Count == 0) {
            Thread.Sleep(50);
            continue;
          }

          // send response
          var response = ImapEncoding.Transfer8Bit.GetBytes(responseQueue.Dequeue());

          try {
            stream.Write(response, 0, response.Length);
          }
          catch (IOException ex) {
            Console.WriteLine("failed at send, client disconnected?");
            Console.WriteLine(ex);
            return;
          }

          // receive request
          var buffer = new byte[0x1000];

          for (;;) {
            try {
              var read = stream.Read(buffer, 0, buffer.Length);

              var lastIndex = 0;
              var cr = false;
              var index = 0;

              for (index = 0; index < read; index++) {
                if ((cr && buffer[index] == 0x0a) || index + 1 == read) {
                  // LF
                  lock ((requestQueue as System.Collections.ICollection).SyncRoot) {
                    requestQueue.Enqueue(ImapEncoding.Transfer8Bit.GetString(buffer, lastIndex, index - lastIndex + 1));
                  }

                  lastIndex = index + 1;
                  cr = false;
                }
                else {
                  cr = (buffer[index] == 0x0d);
                }
              }

              if (read < buffer.Length)
                break;
            }
            catch (IOException ex) {
              Console.WriteLine("failed at receive, client disconnected?");
              Console.WriteLine(ex);
              return;
            }
          }
        }
      }
      catch (ThreadAbortException) {
        client.Close();
      }
    }

    private TcpListener server;
    private Thread thread;
    private readonly Queue<string> responseQueue = new Queue<string>();
    private readonly Queue<string> requestQueue = new Queue<string>();
  }
}
