// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

namespace Smdn.Interop.Sspi {
  [CLSCompliant(false)]
  public class SspiClient : SspiBase {
    public SspiClient(string targetName)
    {
      this.targetName = targetName;
    }

    protected override void Dispose(bool disposing)
    {
      if (hasContext) {
        secur32.DeleteSecurityContext(ref hCtxt);
        hasContext = false;
      }
    }

    protected override SECURITY_STATUS InitializeSecurityContext(CredHandle hCred, byte[] inputToken, out byte[] outputToken)
    {
      var status = SECURITY_STATUS.SEC_E_OK;

      outputToken = SecBufferDesc.ProcessBuffer(delegate(SecBufferDesc outputTokenBufferDesc) {
        var fContextReq = ISC_REQ.ISC_REQ_MUTUAL_AUTH | ISC_REQ.ISC_REQ_STREAM | ISC_REQ.ISC_REQ_CONFIDENTIALITY | ISC_REQ.ISC_REQ_ALLOCATE_MEMORY;

        if (!hasContext) {
          // first call
          status = secur32.InitializeSecurityContext(ref hCred,
                                                     IntPtr.Zero,
                                                     targetName,
                                                     (ISC_REQ)fContextAttr | fContextReq,
                                                     0,
                                                     secur32.SECURITY_NETWORK_DREP,
                                                     IntPtr.Zero,
                                                     0,
                                                     out hCtxt,
                                                     out outputTokenBufferDesc,
                                                     out fContextAttr,
                                                     out tsExpiry);
        }
        else {
          SecBufferDesc.ProcessBuffer(delegate(SecBufferDesc inputTokenBufferDesc) {
            status = secur32.InitializeSecurityContext(ref hCred,
                                                       ref hCtxt,
                                                       targetName,
                                                       (ISC_REQ)fContextAttr | fContextReq,
                                                       0,
                                                       secur32.SECURITY_NETWORK_DREP,
                                                       ref inputTokenBufferDesc,
                                                       0,
                                                       out hCtxt,
                                                       out outputTokenBufferDesc,
                                                       out fContextAttr,
                                                       out tsExpiry);
          });
        }

        if (0x80000000 <= (uint)status)
          throw new SspiException(status);

        hasContext = true;

        if (status == SECURITY_STATUS.SEC_I_COMPLETE_NEEDED || status == SECURITY_STATUS.SEC_I_COMPLETE_AND_CONTINUE)
          // secur32.CompleteAuthToken(ref base.hCtxt, ref outTokenBufferDesc)
          throw new NotSupportedException();
      });

      return status;
    }

    private string targetName;
    private bool hasContext = false;
    private CtxtHandle hCtxt;
    private uint fContextAttr = 0;
  }
}
