// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

namespace Smdn.Interop.Sspi {
  [CLSCompliant(false)]
  public class SspiException : SystemException {
    public SECURITY_STATUS Status {
      get; private set;
    }

    public static void ThrowIfError(SECURITY_STATUS status)
    {
      if ((uint)status < 0x80000000)
        return;

      throw new SspiException(status);
    }

    public SspiException(SECURITY_STATUS status)
      : base(ToString(status))
    {
      this.Status = status;
    }

    private static string ToString(SECURITY_STATUS status)
    {
      if (Enum.IsDefined(typeof(SECURITY_STATUS), status))
        return status.ToString();
      else
        return string.Format("unknown error: 0x{0:x}", status);
    }
  }
}
