// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2009 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Runtime.InteropServices;

namespace Smdn.Interop.Sspi {
  [CLSCompliant(false)]
  public enum SECPKG_CRED : uint {
    SECPKG_CRED_INBOUND   = 0x00000001,
    SECPKG_CRED_OUTBOUND  = 0x00000002,
    SECPKG_CRED_BOTH      = SECPKG_CRED_INBOUND | SECPKG_CRED_OUTBOUND,
    SECPKG_CRED_DEFAULT   = 0x00000004,
  }

  [CLSCompliant(false)]
  public enum SECBUFFER_TYPE : uint {
    SECBUFFER_EMPTY                   = 0x00000000,
    SECBUFFER_DATA                    = 0x00000001,
    SECBUFFER_TOKEN                   = 0x00000002,
    SECBUFFER_PKG_PARAMS              = 0x00000003,
    SECBUFFER_MISSING                 = 0x00000004,
    SECBUFFER_EXTRA                   = 0x00000005,
    SECBUFFER_STREAM_TRAILER          = 0x00000006,
    SECBUFFER_STREAM_HEADER           = 0x00000007,
    SECBUFFER_MECHLIST                = 0x0000000b,
    SECBUFFER_MECHLIST_SIGNATURE      = 0x0000000c,
    SECBUFFER_TARGET                  = 0x0000000d,
    SECBUFFER_CHANNEL_BINDINGS        = 0x0000000e,
    SECBUFFER_CHANGE_PASS_RESPONSE    = 0x0000000f,
    SECBUFFER_TARGET_HOST             = 0x00000010,
    SECBUFFER_ALERT                   = 0x00000011,

    SECBUFFER_READONLY_WITH_CHECKSUM  = 0x10000000,
    SECBUFFER_READONLY                = 0x80000000,
    SECBUFFER_ATTRMASK                = 0xf0000000,
  }

  [CLSCompliant(false), Flags]
  public enum ISC_REQ : uint {
    None                        = 0,
    ISC_REQ_DELEGATE            = 1 <<  0,
    ISC_REQ_MUTUAL_AUTH         = 1 <<  1,
    ISC_REQ_REPLAY_DETECT       = 1 <<  2,
    ISC_REQ_SEQUENCE_DETECT     = 1 <<  3,
    ISC_REQ_CONFIDENTIALITY     = 1 <<  4,
    ISC_REQ_USE_SESSION_KEY     = 1 <<  5,
    ISC_REQ_PROMPT_FOR_CREDS    = 1 <<  6,
    ISC_REQ_USE_SUPPLIED_CREDS  = 1 <<  7,
    ISC_REQ_ALLOCATE_MEMORY     = 1 <<  8,
    ISC_REQ_USE_DCE_STYLE       = 1 <<  9,
    ISC_REQ_DATAGRAM            = 1 << 10,
    ISC_REQ_CONNECTION          = 1 << 11,
    ISC_REQ_CALL_LEVEL          = 1 << 12,
    ISC_REQ_FRAGMENT_SUPPLIED   = 1 << 13,
    ISC_REQ_EXTENDED_ERROR      = 1 << 14,
    ISC_REQ_STREAM              = 1 << 15,
    ISC_REQ_INTEGRITY           = 1 << 16,
    ISC_REQ_IDENTIFY            = 1 << 17,
    ISC_REQ_NULL_SESSION        = 1 << 18,
    ISC_REQ_MANUAL_CRED_VALIDATION = 1 << 19,
    ISC_REQ_RESERVED1           = 1 << 20,
    ISC_REQ_FRAGMENT_TO_FIT     = 1 << 21,
  }

  [CLSCompliant(false), Flags]
  public enum ASC_REQ : uint {
    None                        = 0,
    ASC_REQ_DELEGATE            = 1 <<  0,
    ASC_REQ_MUTUAL_AUTH         = 1 <<  1,
    ASC_REQ_REPLAY_DETECT       = 1 <<  2,
    ASC_REQ_SEQUENCE_DETECT     = 1 <<  3,
    ASC_REQ_CONFIDENTIALITY     = 1 <<  4,
    ASC_REQ_USE_SESSION_KEY     = 1 <<  5,
    //
    //
    ASC_REQ_ALLOCATE_MEMORY     = 1 <<  8,
    ASC_REQ_USE_DCE_STYLE       = 1 <<  9,
    ASC_REQ_DATAGRAM            = 1 << 10,
    ASC_REQ_CONNECTION          = 1 << 11,
    ASC_REQ_CALL_LEVEL          = 1 << 12,
    //
    //
    ASC_REQ_EXTENDED_ERROR      = 1 << 15,
    ASC_REQ_STREAM              = 1 << 16,
    ASC_REQ_INTEGRITY           = 1 << 17,
    ASC_REQ_LICENSING           = 1 << 18,
    ASC_REQ_IDENTIFY            = 1 << 19,
    ASC_REQ_ALLOW_NULL_SESSION  = 1 << 20,
  }

  [CLSCompliant(false), StructLayout(LayoutKind.Sequential, Pack = 0)]
  public struct SecPkgInfo {
    public uint   fCapabilities;
    public ushort wVersion;
    public ushort wRPCID;
    public uint   cbMaxToken;
    private IntPtr _Name;
    private IntPtr _Comment;

    public static readonly uint Size = (uint)Marshal.SizeOf(typeof(SecPkgInfo));

    public string Name {
      get { return Marshal.PtrToStringUni(_Name); }
    }

    public string Comment {
      get { return Marshal.PtrToStringUni(_Comment); }
    }

    public static SecPkgInfo QuerySecurityPackageInfo(string packageName)
    {
      var pPackageInfo = IntPtr.Zero;

      try {
        SspiException.ThrowIfError(secur32.QuerySecurityPackageInfo(packageName, out pPackageInfo));

        return (SecPkgInfo)Marshal.PtrToStructure(pPackageInfo, typeof(SecPkgInfo));
      }
      finally {
        if (pPackageInfo != IntPtr.Zero)
          secur32.FreeContextBuffer(pPackageInfo);
      }
    }
  }

  [CLSCompliant(false), StructLayout(LayoutKind.Sequential)]
  public struct SecHandle {
    public static readonly SecHandle Zero = new SecHandle(0, 0);

    public uint dwLower;
    public uint dwUpper;

    private SecHandle(uint lower, uint upper)
    {
      dwLower = lower;
      dwUpper = upper;
    }
  }

  [CLSCompliant(false), StructLayout(LayoutKind.Sequential)]
  public struct CtxtHandle {
    public static readonly CtxtHandle Zero = new CtxtHandle(0, 0);

    public uint dwLower;
    public uint dwUpper;

    private CtxtHandle(uint lower, uint upper)
    {
      dwLower = lower;
      dwUpper = upper;
    }
  }

  [CLSCompliant(false), StructLayout(LayoutKind.Sequential)]
  public struct CredHandle {
    public static readonly CredHandle Zero = new CredHandle(0, 0);

    public uint dwLower;
    public uint dwUpper;

    private CredHandle(uint lower, uint upper)
    {
      dwLower = lower;
      dwUpper = upper;
    }
  }

  [CLSCompliant(false), StructLayout(LayoutKind.Sequential)]
  public struct TimeStamp {
    public uint dwLowDateTime;
    public uint dwHighDateTime;

    private TimeStamp(uint low, uint high)
    {
      dwLowDateTime = low;
      dwHighDateTime = high;
    }
  }

  [CLSCompliant(false), StructLayout(LayoutKind.Sequential)]
  public struct SecBuffer {
    public static readonly int Size = Marshal.SizeOf(typeof(SecBuffer));

    public uint cbBuffer;
    public SECBUFFER_TYPE BufferType;
    public IntPtr pvBuffer;
  }

  [CLSCompliant(false), StructLayout(LayoutKind.Sequential)]
  public struct SecBufferDesc {
    public const uint SECBUFFER_VERSION = 0;

    public uint ulVersion;
    public uint cBuffers;
    public IntPtr pBufferes; // PSecBuffer pBuffers

    public static byte[] ProcessBuffer(Action<SecBufferDesc> action)
    {
      unsafe {
        var secBuffer = stackalloc SecBuffer[1];

        secBuffer[0].BufferType = SECBUFFER_TYPE.SECBUFFER_TOKEN;
        secBuffer[0].cbBuffer = (uint)0;
        secBuffer[0].pvBuffer = IntPtr.Zero; // will be allocated by ISC_REQ_ALLOCATE_MEMORY

        var secBufferDesc = new SecBufferDesc();

        secBufferDesc.ulVersion = SECBUFFER_VERSION;
        secBufferDesc.cBuffers = 1;
        secBufferDesc.pBufferes = (IntPtr)secBuffer;

        try {
          action(secBufferDesc);

          if (0 <= secBuffer[0].cbBuffer && secBuffer[0].pvBuffer != IntPtr.Zero) {
            var ret = new byte[secBuffer[0].cbBuffer];

            Marshal.Copy((*secBuffer).pvBuffer, ret, 0, ret.Length);

            return ret;
          }
          else {
            return null;
          }
        }
        finally {
          if ((*secBuffer).pvBuffer != IntPtr.Zero)
            secur32.FreeContextBuffer((*secBuffer).pvBuffer);
        }
      }
    }

    public static void ProcessBuffer(byte[] data, Action<SecBufferDesc> action)
    {
      unsafe {
        var secBuffer = stackalloc SecBuffer[1];

        fixed (void* dataPtr = data) {
          secBuffer[0].BufferType = SECBUFFER_TYPE.SECBUFFER_TOKEN;
          secBuffer[0].cbBuffer = (uint)data.Length;
          secBuffer[0].pvBuffer = (IntPtr)dataPtr;

          var secBufferDesc = new SecBufferDesc();

          secBufferDesc.ulVersion = SecBufferDesc.SECBUFFER_VERSION;
          secBufferDesc.cBuffers = 1;
          secBufferDesc.pBufferes = (IntPtr)secBuffer;

          action(secBufferDesc);
        }
      }
    }
  }
}
